package datavisualizer;

import javax.swing.*;
import javawrapper.ModelException;
import org.jfree.chart.*;

/**
 * This class manages the data for, and draws, tree maps.  It will draw the
 * trees as circles proportional to the dbh.  Seedlings are not drawn.
 *
 * X, Y, and DBH are required to draw a tree map.  While it would be possible
 * to draw a tree map without any kind of DBH or with height instead
 * (back-calculating the dbh), this has not been implemented.
 *
 * Copyright: Copyright (c) Charles D. Canham 2003
 * Company: Institute of Ecosystem Studies
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 * <br>November 12, 2004:  Added features for view (LEM)
 */

public class TreeMapDataRequest
    extends DataRequest
    implements java.awt.event.ActionListener {

  /**These hold the XYZ data (Z being dbh)*/
  private XYZSeriesCollection m_oDataSet;
  private XYZSeries[] mp_oSeries; /**<One for each species*/

  /**Field displaying the DBH scale factor*/
  private JTextField m_jDBHScale = new JTextField("");

  /**<There will be one of these for each species*/
  /**Holds the data codes for X.  Sized # species by # types.*/
  private int[][] mp_iXCode,
      /**Holds the data codes for Y.  Sized # species by # types.*/
      mp_iYCode,
      /**Holds the data codes for DBH.  Sized # species by # types.*/
      mp_iDbhCode;

  /**Tree X coordinate.*/
  private float m_fX,
      /**Tree Y coordinate.*/
      m_fY,
      /**Tree DBH value.*/
      m_fDbh,
      /**Minimum dbh to draw*/
      m_fMinDbh,
      /**DBH scale factor*/
      m_fDBHScale = (float) 0.25;
  private int m_iNumSpecies, /**<Total number of species.*/
      m_iNumTypes; /**<Total number of tree types.*/
  private ChartPanel m_oChartPanel = null; /**<Chart being displayed*/

  private final static int LEFT = 1; /**<Scrolling left*/
  private final static int RIGHT = 2; /**<Scrolling right*/
  private final static int UP = 3; /**<Scrolling up*/
  private final static int DOWN = 4; /**<Scrolling down*/

  /**
   * Constructor
   * @param oManager Parent detailed output file manager
   * @param sChartName Chart name
   * @throws ModelException Passing through possible underlying exceptions
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public TreeMapDataRequest(DetailedOutputFileManager oManager, String sChartName) throws
      ModelException {
    super(sChartName, oManager);
    int i, j;

    m_fX = -1;
    m_fY = -1;
    m_fDbh = -1;
    m_fMinDbh = 5; //default value
    m_iNumSpecies = oManager.GetNumberOfSpecies();
    m_iNumTypes = oManager.GetNumberOfTypes();

    //For now - don't display the "show all timesteps" option - because
    //the default behavior, which is putting all the data in one file,
    //is very bad.  We'll need to come back and make specialized behavior
    //for tree maps.
    m_bShowOneTimestep = false;

    m_oDataSet = new XYZSeriesCollection();
    mp_oSeries = new XYZSeries[m_iNumSpecies];
    for (i = 0; i < m_iNumSpecies; i++) {
      mp_oSeries[i] = new XYZSeries(m_oManager.GetSpeciesNameFromCode(i));
    }

    mp_iXCode = new int[m_iNumSpecies][];
    mp_iYCode = new int[m_iNumSpecies][];
    mp_iDbhCode = new int[m_iNumSpecies][];
    for (i = 0; i < m_iNumSpecies; i++) {
      mp_iXCode[i] = new int[m_iNumTypes];
      mp_iYCode[i] = new int[m_iNumTypes];
      mp_iDbhCode[i] = new int[m_iNumTypes];

      for (j = 0; j < m_iNumTypes; j++) {
        mp_iXCode[i][j] = -1;
        mp_iYCode[i][j] = -1;
        mp_iDbhCode[i][j] = -1;
      }
    }
  }

  /**
   * Writes the tree map's data to tab-delimited text.
   * @param jOut java.io.FileWriter The file to write to.
   * @throws IOException if there is a problem writing the file.
   * @throws ModelException if there's a problem assembling the data.
   */
  protected void WriteChartDataToFile(java.io.FileWriter jOut) throws java.io.
      IOException, javawrapper.ModelException {

    if (mp_oSeries == null) {
      jOut.write("This map has no data.");
      return;
    }

    //Write timestep
    DetailedOutputLegend oLegend = (DetailedOutputLegend) m_oManager.GetLegend();
    jOut.write("Timestep\t" + String.valueOf(oLegend.GetCurrentTimestep()) + "\n");

    //Write header
    jOut.write("X\tY\tDBH\tSpecies\n");

    //Write the data
    int iSp, i;
    XYZDataItem oItem;
    for (iSp = 0; iSp < mp_oSeries.length; iSp++) {
      String sSpeciesName = m_oManager.GetLegend().GetSpeciesDisplayName(iSp);
      int iNumItems = mp_oSeries[iSp].getItemCount();
      for (i = 0; i < iNumItems; i++) {
        oItem = mp_oSeries[iSp].getDataItem(i);
        //X
        jOut.write(oItem.getX().toString());
        //Y
        jOut.write("\t" + oItem.getY().toString());
        //DBH
        jOut.write("\t" + oItem.getZ().toString());
        //Species
        jOut.write("\t" + sSpeciesName + "\n");
      }
    }
  }

  /**
   * Sets the minimum dbh value to show on the map.
   * @param f The new minimum.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetMinimumDbh(float f) {
    m_fMinDbh = f;
  }

  /**
   * Gets the minimum dbh value to show on the map.
   * @return The minimum dbh value.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public float GetMinimumDbh() {
    return m_fMinDbh;
  }

  public boolean WantAnyTreeFloats() {
    return true;
  }

  void ClearData() throws ModelException {
    //Garbage-collect and redeclare everything
    m_oDataSet = null;
    m_oDataSet = new XYZSeriesCollection();

    for (int i = 0; i < m_iNumSpecies; i++) {
      mp_oSeries[i] = null;
      mp_oSeries[i] = new XYZSeries(m_oManager.GetSpeciesNameFromCode(i));
    }
  }

  public void AddTreeFloatData(int iSpecies, int iType, int iCode, float fVal) {

    if (iType <= javawrapper.TreePopulation.SEEDLING) {
      return;
    }

    if (iCode == mp_iXCode[iSpecies][iType]) {
      m_fX = fVal;
    }
    else if (iCode == mp_iYCode[iSpecies][iType]) {
      m_fY = fVal;
    }
    else if (iCode == mp_iDbhCode[iSpecies][iType]) {
      m_fDbh = fVal;
    }

    if (m_fX >= 0 && m_fY >= 0 && m_fDbh >= 0) {
      if (m_fDbh >= m_fMinDbh) {
        mp_oSeries[iSpecies].add(m_fX, m_fY, m_fDbh);
      }
      m_fX = -1;
      m_fY = -1;
      m_fDbh = -1;
    }
  }

  public void AddTreeFloatDataMemberCode(int iSpecies, int iType, String sLabel,
                                         int iCode) {

    if (sLabel.equalsIgnoreCase("x")) {
      mp_iXCode[iSpecies][iType] = iCode;
    }
    else if (sLabel.equalsIgnoreCase("y")) {
      mp_iYCode[iSpecies][iType] = iCode;
    }
    else if (sLabel.equalsIgnoreCase("dbh")) {
      mp_iDbhCode[iSpecies][iType] = iCode;
    }
  }

  /**
   * Creates the tree map window.
   * @param oLegend Legend Legend for this detailed output file.
   * @param sChartTitle String Title of chart.
   * @throws ModelException if there is a problem drawing the window.
   * @return JInternalFrame Frame with tree map.
   */
  JInternalFrame DrawChart(Legend oLegend, String sChartTitle) throws
      ModelException {
    //Add the series to the dataset
    int i;
    for (i = 0; i < mp_oSeries.length; i++) {
      if (mp_oSeries[i].getItemCount() > 0) {
        m_oDataSet.addSeries(mp_oSeries[i]);
      }
    }

    //Now adjust so that only those visible species are displayed by making a
    //copy
    XYZSeriesCollection oAdjustedDataset = UpdateForVisible(oLegend);

    m_oChartFrame = DataGrapher.DrawTreeMap(oAdjustedDataset, "East ->",
                                            "",
                                            sChartTitle,
                                            (int) m_oManager.
                                            GetXPlotLength(),
                                            (int) m_oManager.
                                            GetYPlotLength(), m_fDBHScale,
                                            this, oLegend);

    //Grab the chart panel
    m_oChartPanel = (ChartPanel) DataGrapher.FindNamedComponent(m_oChartFrame.
        getContentPane(), "tree_map_panel");

    //Make a panel for the restore button and the DBH scale
    JPanel jButtonPanel = new JPanel(new java.awt.FlowLayout(java.awt.
        FlowLayout.
        LEFT));
    JButton jRestoreButton = new JButton("Restore Original View");
    jRestoreButton.setFont(new javawrapper.ModelFont());
    jRestoreButton.setActionCommand("Restore");
    jRestoreButton.addActionListener(this);
    JLabel jTempLabel = new JLabel("DBH Scale Factor:");
    jTempLabel.setFont(new javawrapper.ModelFont());
    m_jDBHScale.setFont(new javawrapper.ModelFont());
    m_jDBHScale.setText(String.valueOf(m_fDBHScale));
    JButton jChangeButton = new JButton("Update Map");
    jChangeButton.setFont(new javawrapper.ModelFont());
    jChangeButton.setActionCommand("ChangeScale");
    jChangeButton.addActionListener(this);
    jButtonPanel.add(jRestoreButton);
    jButtonPanel.add(jTempLabel);
    jButtonPanel.add(m_jDBHScale);
    jButtonPanel.add(jChangeButton);

    //Make a pane for scrolling around the map
    JPanel jMapPanel = new JPanel(new java.awt.BorderLayout());
    JButton jUpButton = new JButton(new javawrapper.ModelIcon(15, 15,
        javawrapper.ModelIcon.UP_TRIANGLE, java.awt.Color.BLACK));
    jUpButton.addActionListener(this);
    jUpButton.setActionCommand("ScrollUp");
    jUpButton.setToolTipText("Scroll north");
    JButton jDownButton = new JButton(new javawrapper.ModelIcon(15, 15,
        javawrapper.ModelIcon.DOWN_TRIANGLE, java.awt.Color.BLACK));
    jDownButton.addActionListener(this);
    jDownButton.setActionCommand("ScrollDown");
    jDownButton.setToolTipText("Scroll south");
    JButton jLeftButton = new JButton(new javawrapper.ModelIcon(15, 15,
        javawrapper.ModelIcon.LEFT_TRIANGLE, java.awt.Color.BLACK));
    jLeftButton.addActionListener(this);
    jLeftButton.setActionCommand("ScrollLeft");
    jLeftButton.setToolTipText("Scroll west");
    JButton jRightButton = new JButton(new javawrapper.ModelIcon(15, 15,
        javawrapper.ModelIcon.RIGHT_TRIANGLE, java.awt.Color.BLACK));
    jRightButton.addActionListener(this);
    jRightButton.setActionCommand("ScrollRight");
    jRightButton.setToolTipText("Scroll east");
    //Make all the buttons the same thickness
    int iHeight = (int) jUpButton.getPreferredSize().getHeight();
    jLeftButton.setPreferredSize(new java.awt.Dimension(iHeight,
        (int) jLeftButton.getPreferredSize().getHeight()));
    jRightButton.setPreferredSize(new java.awt.Dimension(iHeight,
        (int) jRightButton.getPreferredSize().getHeight()));

    jMapPanel.add(m_oChartPanel, java.awt.BorderLayout.CENTER);
    jMapPanel.add(jUpButton, java.awt.BorderLayout.NORTH);
    jMapPanel.add(jLeftButton, java.awt.BorderLayout.WEST);
    jMapPanel.add(jRightButton, java.awt.BorderLayout.EAST);
    jMapPanel.add(jDownButton, java.awt.BorderLayout.SOUTH);

    JPanel jContentPanel = new JPanel(new java.awt.BorderLayout());
    jContentPanel.add(jButtonPanel, java.awt.BorderLayout.NORTH);
    jContentPanel.add(jMapPanel, java.awt.BorderLayout.CENTER);
    m_oChartFrame.setContentPane(jContentPanel);

    return m_oChartFrame;
  }

  /**
   * Updates the tree map.
   * @param oLegend Legend Legend for this detailed output file.
   * @throws ModelException if there is a problem drawing the window.
   */
  public void UpdateChart(Legend oLegend) throws ModelException {

    //If there are no series in the dataset, add them now (in case the
    //file was re-parsed while this chart was open)
    if (m_oDataSet.getSeriesCount() == 0) {
      int i;
      for (i = 0; i < mp_oSeries.length; i++) {
        if (mp_oSeries[i].getItemCount() > 0) {
          m_oDataSet.addSeries(mp_oSeries[i]);
        }
      }
    }

    //Adjust our dataset so that only those visible species are displayed
    XYZSeriesCollection oAdjustedDataset = UpdateForVisible(oLegend);

    DataGrapher.UpdateTreeMap(oAdjustedDataset, m_oChartFrame, oLegend,
                              m_fDBHScale);

    //Grab the chart panel
    m_oChartPanel = (ChartPanel) DataGrapher.FindNamedComponent(m_oChartFrame.
        getContentPane(), "tree_map_panel");
  }

  /**
   * Returns a copy of the dataset with only those species which are marked as
   * visible in the legend.
   * @param oLegend The legend for this chart.
   * @return The copied dataset.
   * @throws ModelException Passing through underlying exceptions.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected XYZSeriesCollection UpdateForVisible(Legend oLegend) throws
      ModelException {
    XYZSeriesCollection oAdjustedDataset = new XYZSeriesCollection();
    XYZSeries oSeries;
    String sName;
    int i;

    for (i = 0; i < m_oDataSet.getSeriesCount(); i++) {
      oSeries = (XYZSeries) m_oDataSet.getSeries(i);

      //What species is this for?
      sName = oSeries.getName();

      //If it's enabled, add the series to the copy
      if (oLegend.GetIsSpeciesSelected(m_oManager.GetSpeciesCodeFromName(sName))) {
        oAdjustedDataset.addSeries(oSeries);
      }
    }
    return oAdjustedDataset;
  }

  /**
   * Restores the original view of the tree map with the scale at which it
   * was first displayed.
   */
  protected void RestoreOriginalView() {
    if (m_oChartPanel == null) {
      return;
    }

    org.jfree.chart.axis.NumberAxis oXAxis = (org.jfree.chart.axis.NumberAxis)
        m_oChartPanel.getChart().getXYPlot().getDomainAxis();
    oXAxis.setLowerBound(0);
    oXAxis.setUpperBound( (double) m_oManager.GetXPlotLength());

    org.jfree.chart.axis.NumberAxis oYAxis = (org.jfree.chart.axis.NumberAxis)
        m_oChartPanel.getChart().getXYPlot().getRangeAxis();
    oYAxis.setLowerBound(0);
    oYAxis.setUpperBound( (double) m_oManager.GetYPlotLength());
  }

  /**
   * Scrolls the view 10% in the specified direction.  If the view is full-size
   * nothing happens.  If the view is already scrolled all the way over in the
   * specified direction nothing happens.
   * @param iDirection int Direction in which to scroll.
   */
  protected void Scroll(int iDirection) {

    if (m_oChartPanel == null) {
      return;
    }

    double fUpMax = (double) m_oManager.GetYPlotLength(),
        fRightMax = (double) m_oManager.GetXPlotLength(),
        fDownMax = 0, fLeftMax = 0, fNewMin = 0, fNewMax = 0, fDirectionInc,
        fCurrentMin, fCurrentMax;

    org.jfree.chart.axis.NumberAxis oXAxis = (org.jfree.chart.axis.NumberAxis)
        m_oChartPanel.getChart().getXYPlot().getDomainAxis();
    org.jfree.chart.axis.NumberAxis oYAxis = (org.jfree.chart.axis.NumberAxis)
        m_oChartPanel.getChart().getXYPlot().getRangeAxis();
    org.jfree.chart.axis.NumberAxis oAxis;

    if (iDirection != LEFT && iDirection != RIGHT && iDirection != UP &&
        iDirection != DOWN) {
      return;
    }

    //Get the current bounds
    if (iDirection == LEFT || iDirection == RIGHT) {
      oAxis = oXAxis;
    }
    else {
      oAxis = oYAxis;
    }

    fCurrentMin = oAxis.getLowerBound();
    fCurrentMax = oAxis.getUpperBound();
    fDirectionInc = (fCurrentMax - fCurrentMin) / 10;

    //Exit if we're already scrolled
    if (iDirection == LEFT) {
      if (fCurrentMin == fLeftMax) {
        return;
      }

      fNewMin = java.lang.Math.max(fCurrentMin - fDirectionInc, fLeftMax);
      fNewMax = fCurrentMax - fDirectionInc;
    }
    else if (iDirection == RIGHT) {
      if (fCurrentMax == fRightMax) {
        return;
      }

      fNewMin = fCurrentMin + fDirectionInc;
      fNewMax = java.lang.Math.min(fCurrentMax + fDirectionInc, fRightMax);

    }
    else if (iDirection == UP) {
      if (fCurrentMax == fUpMax) {
        return;
      }

      fNewMin = fCurrentMin + fDirectionInc;
      fNewMax = java.lang.Math.min(fCurrentMax + fDirectionInc, fUpMax);
    }
    else if (iDirection == DOWN) {
      if (fCurrentMin == fDownMax) {
        return;
      }

      fNewMin = java.lang.Math.max(fCurrentMin - fDirectionInc, fDownMax);
      fNewMax = fCurrentMax - fDirectionInc;
    }

    oAxis.setLowerBound(fNewMin);
    oAxis.setUpperBound(fNewMax);
  }

  /**
   * Performs actions for the controls in the Histogram window.
   * @param oEvent Event triggered.
   */
  public void actionPerformed(java.awt.event.ActionEvent oEvent) {
    String sCommand = oEvent.getActionCommand();
    if (sCommand.equals("Restore")) {
      RestoreOriginalView();
    }
    else if (sCommand.equals("ScrollUp")) {
      Scroll(UP);
    }
    else if (sCommand.equals("ScrollDown")) {
      Scroll(DOWN);
    }
    else if (sCommand.equals("ScrollLeft")) {
      Scroll(LEFT);
    }
    else if (sCommand.equals("ScrollRight")) {
      Scroll(RIGHT);
    }
    else if (sCommand.equals("ChangeScale")) {

      //Get the value in the text field
      float fNewScale;
      try {
        Float oNewScale = new Float(m_jDBHScale.getText());
        fNewScale = oNewScale.floatValue();
      }
      catch (java.lang.NumberFormatException oErr) {
        JOptionPane.showMessageDialog(m_oChartFrame,
                                      "The value for DBH scale is not a recognized number.");
        m_jDBHScale.setText(String.valueOf(m_fDBHScale));
        return;
      }

      if (fNewScale <= 0) {
        JOptionPane.showMessageDialog(m_oChartFrame,
                                      "The values for DBH scale must be greater than zero.");
        m_jDBHScale.setText(String.valueOf(m_fDBHScale));
        return;
      }
      try {
        m_fDBHScale = fNewScale;
        UpdateChart(m_oManager.GetLegend());
      }
      catch (javawrapper.ModelException oErr) {
        javawrapper.ErrorGUI oHandler = new javawrapper.ErrorGUI(m_oChartFrame);
        oHandler.WriteErrorMessage(oErr);
      }

    }
    super.actionPerformed(oEvent);
  }
}
